
#include <csignal>
#include <cstring>
#include <sstream>
#include <string>
#include <thread>

#include "common/file_parser/helper_h264_parser.h"
#include "common/helper.h"
#include "common/opt_parser.h"
#include "common/sample_common.h"
#include "common/sample_connection_observer.h"
#include "common/sample_local_user_observer.h"
#include "common/switch_video_stream_base.h"
#include "common/log.h"

#define DEFAULT_CONNECT_TIMEOUT_MS (3000)
#define DEFAULT_SAMPLE_RATE (48000)
#define DEFAULT_NUM_OF_CHANNELS (1)
#define DEFAULT_FRAME_RATE_1 (30)
#define DEFAULT_FRAME_RATE_2 (30)
#define DEFAULT_AUDIO_FILE "test_data/test.pcm"
#define DEFAULT_VIDEO_FILE_1 "test_data/test_multi_slice.h264"
#define DEFAULT_VIDEO_FILE_2 "test_data/test_multi_slice.h264"
#define DEFAULT_RECEIVE_DATASTREAM_INFO "start"

static std::string required_receive_datastream =
    DEFAULT_RECEIVE_DATASTREAM_INFO;
static uint64_t FrameCount = 0;

struct SampleOptions {
  std::string appId;
  std::string channelId;
  std::string userId;
  std::string audioFile = DEFAULT_AUDIO_FILE;
  std::string videoFile1 = DEFAULT_VIDEO_FILE_1;
  std::string videoFile2 = DEFAULT_VIDEO_FILE_2;
  int switch_videoFile_delay = 0;
  std::string datastream_info = DEFAULT_RECEIVE_DATASTREAM_INFO;
  struct {
    int sampleRate = DEFAULT_SAMPLE_RATE;
    int numOfChannels = DEFAULT_NUM_OF_CHANNELS;
  } audio;
  struct {
    int frameRate = DEFAULT_FRAME_RATE_1;
  } video1;
  struct {
    int frameRate = DEFAULT_FRAME_RATE_2;
  } video2;
  int rsfec_min_level = 1;
  std::string vos_ip;
};

static void sendOnePcmFrame(
    const SampleOptions &options,
    agora::agora_refptr<agora::rtc::IAudioPcmDataSender> audioFrameSender) {
  static FILE *file = nullptr;
  const char *fileName = options.audioFile.c_str();

  // Calculate byte size for 10ms audio samples
  int sampleSize = sizeof(int16_t) * options.audio.numOfChannels;
  int samplesPer10ms = options.audio.sampleRate / 100;
  int sendBytes = sampleSize * samplesPer10ms;

  if (!file) {
    if (!(file = fopen(fileName, "rb"))) {
      AG_LOG(ERROR, "Failed to open audio file %s", fileName);
      return;
    }
    AG_LOG(INFO, "Open audio file %s successfully", fileName);
  }

  uint8_t frameBuf[sendBytes];

  if (fread(frameBuf, 1, sizeof(frameBuf), file) != sizeof(frameBuf)) {
    if (feof(file)) {
      fclose(file);
      file = nullptr;
      AG_LOG(INFO, "End of audio file");
    } else {
      AG_LOG(ERROR, "Error reading audio data: %s", std::strerror(errno));
    }
    return;
  }

  if (audioFrameSender->sendAudioPcmData(
          frameBuf, 0, samplesPer10ms, agora::rtc::TWO_BYTES_PER_SAMPLE, options.audio.numOfChannels,
          options.audio.sampleRate) < 0) {
    AG_LOG(ERROR, "Failed to send audio frame!");
  }
}

static void SampleSendAudioTask(
    const SampleOptions &options,
    agora::agora_refptr<agora::rtc::IAudioPcmDataSender> audioFrameSender,
    bool &exitFlag) {
  // Currently only 10 ms PCM frame is supported. So PCM frames are sent at 10
  // ms interval
  PacerInfo pacer = {0, 10, std::chrono::steady_clock::now()};

  while (!exitFlag) {
    sendOnePcmFrame(options, audioFrameSender);
    waitBeforeNextSend(pacer);  // sleep for a while before sending next frame
  }
}

static void sendOneH264Frame(
    int frameRate, std::unique_ptr<HelperH264Frame> h264Frame,
    agora::agora_refptr<agora::rtc::IVideoEncodedImageSender>
        videoH264FrameSender) {
  agora::rtc::EncodedVideoFrameInfo videoEncodedFrameInfo;
  videoEncodedFrameInfo.rotation = agora::rtc::VIDEO_ORIENTATION_0;
  videoEncodedFrameInfo.codecType = agora::rtc::VIDEO_CODEC_H264;
  videoEncodedFrameInfo.framesPerSecond = frameRate;
  videoEncodedFrameInfo.frameType =
      (h264Frame.get()->isKeyFrame
           ? agora::rtc::VIDEO_FRAME_TYPE::VIDEO_FRAME_TYPE_KEY_FRAME
           : agora::rtc::VIDEO_FRAME_TYPE::VIDEO_FRAME_TYPE_DELTA_FRAME);
  ;
  FrameCount++;
  /*   AG_LOG(DEBUG, "sendEncodedVideoImage, buffer %p, len %d, frameType %d",
           reinterpret_cast<uint8_t*>(h264Frame.get()->buffer.get()),
     h264Frame.get()->bufferLen, videoEncodedFrameInfo.frameType); */

  /*printf("[ %s ]--> Now  Send %luth Frame H264 Data,  Send frameRate=%d\n ",
         getCurrentSystemTimeChrono().c_str(), FrameCount, frameRate);*/
  videoH264FrameSender->sendEncodedVideoImage(
      reinterpret_cast<uint8_t *>(h264Frame.get()->buffer.get()),
      h264Frame.get()->bufferLen, videoEncodedFrameInfo);
}

static void SampleSendVideoH264Task(
    const SampleOptions &options,
    agora::agora_refptr<agora::rtc::IVideoEncodedImageSender>
        videoH264FrameSender,
    bool &exitFlag) {
  std::unique_ptr<HelperH264FileParser> h264FileParser_1(
      new HelperH264FileParser(options.videoFile1.c_str()));
  h264FileParser_1->initialize();
  std::unique_ptr<HelperH264FileParser> h264FileParser_2(
      new HelperH264FileParser(options.videoFile2.c_str()));
  h264FileParser_2->initialize();
  // Calculate send interval based on frame rate. H264 frames are sent at this
  // interval
  PacerInfo pacer = {0, 1000 / options.video1.frameRate,
                     std::chrono::steady_clock::now()};

  while (!exitFlag) {
    if (SwitchVideoStreamSourceBase::get_cur_flag() == 0) {
      if (SwitchVideoStreamSourceBase::get_switch_status() == true) {
        FrameCount = 0;
        h264FileParser_1->setFileParseRestart();
        SwitchVideoStreamSourceBase::reset_Switch();
        std::this_thread::sleep_for(
            std::chrono::milliseconds(options.switch_videoFile_delay));
      }

      pacer = {0, 1000 / options.video1.frameRate,
               std::chrono::steady_clock::now()};
      if (auto h264Frame = h264FileParser_1->getH264Frame()) {
        sendOneH264Frame(options.video1.frameRate, std::move(h264Frame),
                         videoH264FrameSender);
        waitBeforeNextSend(
            pacer);  // sleep for a while before sending next frame
      }
    } else if (SwitchVideoStreamSourceBase::get_cur_flag() == 1) {
      if (SwitchVideoStreamSourceBase::get_switch_status() == true) {
        FrameCount = 0;
        h264FileParser_2->setFileParseRestart();
        SwitchVideoStreamSourceBase::reset_Switch();
        std::this_thread::sleep_for(
            std::chrono::milliseconds(options.switch_videoFile_delay));
      }
      pacer = {0, 1000 / options.video2.frameRate,
               std::chrono::steady_clock::now()};
      if (auto h264Frame = h264FileParser_2->getH264Frame()) {
        sendOneH264Frame(options.video2.frameRate, std::move(h264Frame),
                         videoH264FrameSender);
        waitBeforeNextSend(
            pacer);  // sleep for a while before sending next frame
      }
    }
  };
}

class SampleConnectionObserverHandle : public SampleConnectionObserver {
 public:
  SampleConnectionObserverHandle();
  virtual ~SampleConnectionObserverHandle();

 public:
  // inherit from agora::rtc::IRtcConnectionObserver
  void onStreamMessageError(agora::user_id_t userId, int streamId, int code,
                            int missed, int cached) override;

 private:
  std::map<std::pair<unsigned int, int>, DataStreamResult> data_stream_stats;
};

SampleConnectionObserverHandle::SampleConnectionObserverHandle() {}

SampleConnectionObserverHandle::~SampleConnectionObserverHandle() {}

void SampleConnectionObserverHandle::onStreamMessageError(
    agora::user_id_t userId, int streamId, int code, int missed, int cached) {
  printf(
      "*** Receive StreamMessage error, userId = %s, streamId = %d, code = "
      "%d, missed = %d, cached = %d\n",
      userId, streamId, code, missed, cached);
}



class SampleLocalUserObserverHandle : public SampleLocalUserObserver {
public:
  SampleLocalUserObserverHandle(agora::rtc::ILocalUser* local_user);
  virtual ~SampleLocalUserObserverHandle();
public:
  // inherit from agora::rtc::ILocalUserObserver
  void onStreamMessage(agora::user_id_t userId, int streamId, const char *data,
                        size_t length) override;
private:
  std::map<std::pair<unsigned int, int>, DataStreamResult> data_stream_stats;
};

SampleLocalUserObserverHandle::SampleLocalUserObserverHandle(agora::rtc::ILocalUser* local_user)
  :SampleLocalUserObserver(local_user){}
SampleLocalUserObserverHandle::~SampleLocalUserObserverHandle(){}

void SampleLocalUserObserverHandle::onStreamMessage(agora::user_id_t userId,
                                                     int streamId,
                                                     const char *data,
                                                     size_t length) {
  printf(
      "[ %s ]--> Now  Receive a Data Stream Message,  Data stream received: "
      "userId = %s, streamId = %d, length = %d\n",
      getCurrentSystemTimeChrono().c_str(), userId, streamId,
      static_cast<int>(length));
  unsigned int uid = static_cast<unsigned int>(atoll(userId));
  auto key = std::make_pair(uid, streamId);
  if (data_stream_stats.find(key) == data_stream_stats.end()) {
    DataStreamResult result;
    data_stream_stats.insert(
        std::pair<std::pair<unsigned int, int>, DataStreamResult>(key, result));
  }

  if (data == nullptr || length == 0) {
    printf("*** Data stream received fail, data = %p, length = %d\n", data,
           static_cast<int>(length));
    data_stream_stats[key].check_result = false;
    return;
  }
  int req_data_len = required_receive_datastream.size();

  if (strncmp(data, required_receive_datastream.c_str(), req_data_len) != 0) {
    data_stream_stats[key].check_result = false;
    return;
  }
  data_stream_stats[key].received_msg_count++;
  data_stream_stats[key].received_total_bytes += length;
  SwitchVideoStreamSourceBase::exchange_cur_next_flag();
  printf("cur flag is :%d\n", SwitchVideoStreamSourceBase::get_cur_flag());
}


static bool exitFlag = false;
static void SignalHandler(int sigNo) { exitFlag = true; }

int main(int argc, char *argv[]) {
  SampleOptions options;
  opt_parser optParser;

  optParser.add_long_opt("token", &options.appId,
                         "The token for authentication / must");
  optParser.add_long_opt("channelId", &options.channelId, "Channel Id / must");
  optParser.add_long_opt("userId", &options.userId, "User Id / default is 0");
  optParser.add_long_opt("audioFile", &options.audioFile,
                         "The audio file in raw PCM format to be sent");
  optParser.add_long_opt("sampleRate", &options.audio.sampleRate,
                         "Sample rate for the PCM file to be sent");
  optParser.add_long_opt("numOfChannels", &options.audio.numOfChannels,
                         "Number of channels for the PCM file to be sent");
  optParser.add_long_opt("videoFile1", &options.videoFile1,
                         "The video file1 in H264 format to be sent");
  optParser.add_long_opt("videoFile2", &options.videoFile2,
                         "The video file2 in H264 format to be sent");
  optParser.add_long_opt("fps_1", &options.video1.frameRate,
                         "Target frame rate for sending the video1 stream");
  optParser.add_long_opt("fps_2", &options.video2.frameRate,
                         "Target frame rate for sending the video2 stream");
  optParser.add_long_opt("datastream_info", &options.datastream_info,
                         "The content you want to receive datastream");
  optParser.add_long_opt(
      "switch_delay", &options.switch_videoFile_delay,
      "delay time before sending the switching new file video frame ");
  optParser.add_long_opt("rsfec_min_level", &options.rsfec_min_level,
                         "set the fec level [0 | 1 | 2 | 3 ] ");
  optParser.add_long_opt("vos_ip", &options.vos_ip,
                         "set the vos ip:port address ");

  if ((argc <= 1) || !optParser.parse_opts(argc, argv)) {
    std::ostringstream strStream;
    optParser.print_usage(argv[0], strStream);
    std::cout << strStream.str() << std::endl;
    return -1;
  }

  if (options.appId.empty()) {
    AG_LOG(ERROR, "Must provide appId!");
    return -1;
  }

  if (options.channelId.empty()) {
    AG_LOG(ERROR, "Must provide channelId!");
    return -1;
  }

  std::signal(SIGQUIT, SignalHandler);
  std::signal(SIGABRT, SignalHandler);
  std::signal(SIGINT, SignalHandler);

  if (options.datastream_info.size() > 0) {
    required_receive_datastream = options.datastream_info;
  }

  // Create Agora service
  agora::base::IAgoraService *service =
      createAndInitAgoraService(false, true, true);
  if (!service) {
    AG_LOG(ERROR, "Failed to creating Agora service!");
  }

  // Create Agora connection
  // Create Agora connection
  agora::rtc::RtcConnectionConfiguration ccfg;
  ccfg.autoSubscribeAudio = false;
  ccfg.autoSubscribeVideo = false;
  ccfg.audioRecvMediaPacket = true;
  ccfg.videoRecvMediaPacket = true;
  ccfg.clientRoleType = agora::rtc::CLIENT_ROLE_BROADCASTER;
  ccfg.channelProfile = agora::CHANNEL_PROFILE_CLOUD_GAMING;
  agora::agora_refptr<agora::rtc::IRtcConnection> connection =
      service->createRtcConnection(ccfg);
  if (!connection) {
    AG_LOG(ERROR, "Failed to creating Agora connection!");
    return -1;
  }

  // Register connection observer to monitor connection event
  auto connObserver = std::make_shared<SampleConnectionObserverHandle>();
  connection->registerObserver(connObserver.get());

  auto localUserObserver = std::make_shared<SampleLocalUserObserverHandle>(connection->getLocalUser());

  // set vos server ip address
  agora::base::IAgoraParameter *agoraParameter =
      connection->getAgoraParameter();
  if (options.vos_ip.size() > 0) {
    std::string vos_ip_info = "[\"" + options.vos_ip + "\"]";
    agoraParameter->setObject("rtc.vos_list", vos_ip_info.c_str());
  }
  // set fec level
  agoraParameter->setInt("rtc.video.rsfec_min_level", options.rsfec_min_level);

  // Connect to Agora channel
  if (connection->connect(options.appId.c_str(), options.channelId.c_str(),
                          options.userId.c_str())) {
    AG_LOG(ERROR, "Failed to connect to Agora channel!");
    return -1;
  }
  printf("start send h264 and pcm data!\n");
  // Create media node factory
  agora::agora_refptr<agora::rtc::IMediaNodeFactory> factory =
      service->createMediaNodeFactory();
  if (!factory) {
    AG_LOG(ERROR, "Failed to create media node factory!");
  }

  // Create audio data sender
  agora::agora_refptr<agora::rtc::IAudioPcmDataSender> audioFrameSender =
      factory->createAudioPcmDataSender();
  if (!audioFrameSender) {
    AG_LOG(ERROR, "Failed to create audio data sender!");
    return -1;
  }

  // Create audio track
  agora::agora_refptr<agora::rtc::ILocalAudioTrack> customAudioTrack =
      service->createCustomAudioTrack(audioFrameSender);
  if (!customAudioTrack) {
    AG_LOG(ERROR, "Failed to create audio track!");
    return -1;
  }

  // Create video frame sender
  agora::agora_refptr<agora::rtc::IVideoEncodedImageSender> videoFrameSender =
      factory->createVideoEncodedImageSender();
  if (!videoFrameSender) {
    AG_LOG(ERROR, "Failed to create video frame sender!");
    return -1;
  }

  agora::base::SenderOptions option;
  option.ccMode = agora::base::CC_ENABLED;
  // Create video track
  agora::agora_refptr<agora::rtc::ILocalVideoTrack> customVideoTrack =
      service->createCustomVideoTrack(videoFrameSender, option);
  if (!customVideoTrack) {
    AG_LOG(ERROR, "Failed to create video track!");
    return -1;
  }

  // Publish audio & video track
  connection->getLocalUser()->publishAudio(customAudioTrack);
  connection->getLocalUser()->publishVideo(customVideoTrack);

  // Wait until connected before sending media stream
  connObserver->waitUntilConnected(DEFAULT_CONNECT_TIMEOUT_MS);

  // Start sending media data
  AG_LOG(INFO, "Start sending audio & video data ...");
  std::thread sendAudioThread(SampleSendAudioTask, options, audioFrameSender,
                              std::ref(exitFlag));
  std::thread sendVideoThread(SampleSendVideoH264Task, options,
                              videoFrameSender, std::ref(exitFlag));

  sendAudioThread.join();
  sendVideoThread.join();

  // Unpublish audio & video track
  connection->getLocalUser()->unpublishAudio(customAudioTrack);
  connection->getLocalUser()->unpublishVideo(customVideoTrack);

  connection->getLocalUser()->unregisterLocalUserObserver(localUserObserver.get());

  // Disconnect from Agora channel
  bool disconnected = connection->disconnect();
  if (!disconnected) {
    AG_LOG(ERROR, "Failed to disconnect from Agora channel!");
    return -1;
  }

  AG_LOG(INFO, "Disconnected from Agora channel successfully");
  connObserver.reset();
  localUserObserver.reset();
  audioFrameSender = nullptr;
  videoFrameSender = nullptr;
  customAudioTrack = nullptr;
  customVideoTrack = nullptr;
  factory = nullptr;
  connection = nullptr;
  return 0;
}
